/* $Id: et4000.c,v 1.10 1999/06/27 16:07:00 marcus Exp $
******************************************************************************

   ET4000 chipset driver

   Copyright (C) 1997	Andreas Vogler	[duff@allgaeu.org]
   Copyright (C) 1997	Jason McMullan	[jmcc@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

	This code was derived from the following source of information:
	[1]     IBM VGA Driver code by Jason Mcmullan
	[2]     VGADOC 4 by Finn Thoegersen
	[3]	ET6000 Driver code by Jim Ursetto
	[4]	WD Driver code by Brian Julin

******************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Gernot_Ziegler
#define	DRIVER_NAME	"Tseng ET4000 chipset driver"
#define	DRIVER_REV	"$Revision: 1.10 $"

#define SLOW_POINTER 
#define USE_UNTESTED_CODE 1

#define DEBUG_LEVEL 0

#include <kgi/module.h>

#warning This driver uses Linux specific things!!!!!
#ifndef CONFIG_PCI
#warning kernel configured without PCI; W32p support disabled
#endif

enum chip_types 
{
	CHIP_UNKNOWN = -1,
	ET4000 = 0,
	ET4000_W32,
	ET4000_W32i,
	ET4000_W32p
};

char * chipstr[] = {
	"ET4000",
	"ET4000 W32",
	"ET4000 W32i",
	"ET4000 W32p"
};

#ifdef CONFIG_PCI
kgiu16 PCIdev_id[] = {
	PCI_DEVICE_ID_TSENG_W32P_2,
	PCI_DEVICE_ID_TSENG_W32P_b,
	PCI_DEVICE_ID_TSENG_W32P_c,
	PCI_DEVICE_ID_TSENG_W32P_d,
	0
};
#endif

kgi_modereq ggi_auto_textmode = 
{ 1,
	{80,25},
	{80,25},
	{240,200},
	KGIGT_TEXT16,
	{9, 14}
};

kgi_modereq ggi_auto_graphmode = 
{ 1,
	{640,480},
	{640,480},
	{240,200},
	KGIGT_8BIT,
	{1, 1}
};

enum chip_types chip = CHIP_UNKNOWN;

extern kgi_uint law_base;

#include "Tseng/et4000.h"
#include "et4000_io.inc"
#include "et4000_mmio.inc"

static struct kgi_chipset chipinfo =
{
	"Tseng Laboratories",	/* manufacturer         */
	"ET4000",			/* model (autodetect)   */
	GM_TEXT16 |          	/* supported modes - inplement 4 bits ? */
	GM_8BIT | GM_15BIT |	/* CAREFUL ! 24 bit is dangerous ! */
	GM_16BIT | GM_24BIT | GM_32BIT,
	0,
	{2024, 2024},		/* maximum resolution   */
	NULL,			/* Private data hook	*/
  	0x0,			/* autodetect !!! */
	{0, 80000000}		/* DCLK limits            */
};

static struct statetable	/* this table holds the default bits */
 default_state =		/* for text & graphics mode, will be */
{				/* filled in during mode_init        */
	0x23,			/* MISC */
	{
	0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xBC	/* SEQ */
	},
	{
	0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00,		/* CRT */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,
				
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xAB,		/* 0x1x */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x2x */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x23, 0x00,		/* 0x3x */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	},
	{
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* GRC */
	0xFF
	},
	{
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,		/* ATC */
	0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
	0x00, 0x00, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x00
	}
};

struct statetable reset_state;
kgi_modereq default_kgi_modereq;
struct kgi_mode default_kgi_mode;

static unsigned long fb_text_len;
mem_vaddr fb_text;
static unsigned long fb_graph_len;
mem_vaddr fb_graph;

static struct io_region et4000_io = {
	 IO_NULL,
	 0x3C0,	/* VGA IO Region Start */
	 0x20, /* 32 register (0x3C0-0x3DF) */
	 0x3FF,	/* Only 10 address lines are decoded.. */
	 "Tseng ET4000 IO"
};

static struct io_region w32_io = {
	 IO_NULL,
	 0x217A,	/* W32 IO register */
	 0x4,
	 -1, /* Decode all bits */
	 "Tseng W32 IO"
};

static struct mem_region et4000_mem_graph = {
	 MEM_NULL,
	 0xA0000, /* VGA Mem Region Start */
	 0x20000, /* Graphics (128K) region */
	 -1, /* Decode all bits */
	 "Tseng ET4000"
};

static struct mem_region w32_linear_mem = {
	 MEM_NULL,
	 0x0,		/* linear framebuffer (autodetect) */
	 0x400000,	/* 4M */
	 -1,		/* Decode all bits */
	 "Tseng W32 mem"
};

/*
 * Mode specific frame buffers
 */

#include "et4000_text16.inc"
#include "et4000_fb.inc"

/*
 support functions
*/

int tstrg(kgiu16 rg, kgiu8 mask)
{
	kgiu8 old, nw1, nw2;
	
	old=io_in8(rg);
	io_out8(old & ~mask, rg);
	nw1=io_in8(rg) & mask;
	io_out8(old | mask, rg);
	nw2=io_in8(rg) & mask;
	io_out8(old, rg);
	return ((nw1==0) && (nw2==mask));
}

int testCRT(kgiu8 ix, kgiu8 mask)
{
	kgiu8 old, nw1, nw2;
	
	old=inbCRT(ix);
	outbCRT(old & ~mask, ix);
	nw1=inbCRT(ix) & mask;
	outbCRT(old | mask, ix);
	nw2=inbCRT(ix) & mask;
	outbCRT(old, ix);
	return ((nw1==0) && (nw2==mask));
}

int et4000_examine_chipset(void)
{
	kgiu8 val;
	
	/* test for ET4000 family */
	DEBUG1("Detecting chipset...");
	if(tstrg(SEG_SELECT, 0x3F)) {
		if(testCRT(CRT_EXT_ADDR, 0xF)) {
			if(tstrg(EXT_BANK, 0x33)) {
				outbCRT(inbCRT(CRT_VIDEO_CFG1) | 0x28,
					CRT_VIDEO_CFG1);
				val=Sprite_in8(0xEC) >> 4;
				switch(val) {
					
				case 0:
					chip=ET4000_W32;
					break;
					
				case 1:
				case 3:
				case 11:
					chip=ET4000_W32i;
					break;
					
				case 2:
				case 5:
				case 6:
				case 7:
					chip=ET4000_W32p;
					break;
					
				default:
					DEBUG1("unknown ET4000 chip detected: %x", val);
					return -E(CHIPSET, UNKNOWN);
					break;
				}
			} else {
				chip=ET4000;
			}
		} else {
			NOTICE("ET3000 detected, not supported by this driver.");
		}
	}
	if(chip==CHIP_UNKNOWN)
		return -E(CHIPSET, UNKNOWN);
	
	val=inbCRT(CRT_VIDEO_CFG2);
	if(chip==ET4000) {
		chipinfo.ramsize=
			(1 << (val & 3)) * ((val & 8) ? 0x40000 : 0x10000);
	} else {
		chipinfo.ramsize=
			(1 << ((val&1)+1)) * ((val&8) ? 0x40000 : 0x100000);
		if(inbCRT(CRT_RASCAS) & CR32_INTERLEAVE)
			chipinfo.ramsize *= 2;
	}
	strcpy(chipinfo.model, chipstr[chip]);
	DEBUG1("detected chip: %s with %i KByte memory",
	       chipstr[chip], chipinfo.ramsize/1024);
	return EOK;
}

int w32p_examine_chipset(void)
{
	int err,i;
	kgiu8 PCIdevfn, PCIbus;

#ifndef CONFIG_PCI
	ERROR("W32p support disabled");
	return -E(CHIPSET,FAILED);
#else

	/* We assume that there is at least one W32p somewhere */
	/* I am not sure if it would be possible to use more than one */

	if(!pcibios_present()) {
		DEBUG0("Can't access PCI BIOS");
	} else {
		/* just tries to find the first W32p */
		for(i=0; PCIdev_id[i] !=0; i++) {
			if(pcibios_find_device(PCI_VENDOR_ID_TSENG,
					       PCIdev_id[i],
					       0,
					       &PCIbus,
					       &PCIdevfn)
			   == PCIBIOS_SUCCESSFUL)
				break;
		}
		if(PCIdev_id[i] == 0) {
			DEBUG0("could not find W32p in PCI BIOS\n"
			       "This should never happen !!!");
			return -E(CHIPSET, FAILED);
		}
		DEBUG1("found W32p in PCI space: %x, %x",
		       PCIbus,
		       PCIdevfn);
		pcibios_read_config_dword(PCIbus, PCIdevfn,
					  PCI_BASE_ADDRESS_0,
					  &law_base);
		DEBUG1("base address is: %x", law_base);
		
		w32_linear_mem.base=law_base;
		if ((err=mem_check_region(&w32_linear_mem)))
			return err;
		mem_claim_region(&w32_linear_mem);
		DEBUG1("claimed linear memory area");
		mm_linear_graphics.pstart=law_base;
		mm_linear_graphics.psize=chipinfo.ramsize;
	}
	return EOK;
#endif /* CONFIG_PCI */
}

void et4000_release_regions(void)
{
	if(et4000_io.baseptr)
		io_free_region(&et4000_io);
	if(w32_io.baseptr)
		io_free_region(&w32_io);
	if(et4000_mem_graph.baseptr)
		mem_free_region(&et4000_mem_graph);
	if(w32_linear_mem.baseptr)
		mem_free_region(&w32_linear_mem);
}

#ifdef DEBUG
static void printk_mode (struct kgi_mode *mode, char *msg)
{
	#define M(X)	mode->X
	#define R(X)	mode->request.X
	#define T(X)	M(X.width), M(X.blankstart), M(X.syncstart),\
	M(X.syncend), M(X.blankend), M(X.total), M(X.polarity)

		printk(KERN_DEBUG "%s\n",msg);
	printk(KERN_DEBUG "mode = {\n" 
               KERN_DEBUG "   { %i,  /* frames */\n"
               KERN_DEBUG "   { %i, %i},  /* visible */\n",
	       R(frames), R(visible.x),R(visible.y));
	printk(KERN_DEBUG "   { %i, %i},   /* virtual */\n",
	       R(virt.x),R(virt.y));
	printk(KERN_DEBUG "   { %i, %i},   /* size [mm] */\n",
	       R(size.x),R(size.y));
	printk(KERN_DEBUG "   0x%x,\n", R(graphtype));
	printk(KERN_DEBUG "   { %i, %i}},   /* textgrid */\n",
	       R(dpp.x),R(dpp.y));
	printk(KERN_DEBUG "   { %i, %i, %i, %i, %i, %i, %i}, /* x timing */\n",
	       T(x));
	printk(KERN_DEBUG "   { %i, %i, %i, %i, %i, %i, %i}, /* y timing */\n",
	       T(y));
	printk(KERN_DEBUG "   { %i, %i},   /* magnify */\n",
	       M(magnify.x), M(magnify.y));
	printk(KERN_DEBUG "   %x,          /* flags   */\n", M(flags));
	printk(KERN_DEBUG "   %i,          /* dclk    */\n", M(dclk));
	printk(KERN_DEBUG "   { %i, %i},   /* clk     */\n", 
	       M(clk.mul), M(clk.div));
	printk(KERN_DEBUG "   { %i, %i}    /* lclk    */\n"
	       KERN_DEBUG "}\n",
	       M(lclk.mul), M(lclk.div));

	#undef M
	#undef R
	#undef T
}
#endif /* #ifdef DEBUG */

/*
  GGI interface
*/

static void chipset_setup_dpy(struct kgi_display *dpy,struct kgi_mode *mode)
{
	switch (mode->request.graphtype) {
	case KGIGT_TEXT16:
		dpy->set_font=text16_set_font;
		dpy->set_origin=text16_set_origin;
		dpy->set_split=text16_set_split;
		dpy->set_frame=text16_set_frame;
		dpy->show_cursor=text16_show_cursor;
		dpy->hide_cursor=text16_hide_cursor;
		dpy->undo_cursor=text16_undo_cursor;
		dpy->show_pointer=text16_show_pointer;
		dpy->hide_pointer=text16_hide_pointer;
		dpy->undo_pointer=text16_undo_pointer;
		dpy->fb.dpy=dpy;
		dpy->fb.size.x = mode->request.virt.x;
		dpy->fb.size.y = mode->request.virt.y;
		dpy->fb.bpp    = MAP_16_BPP;
		dpy->fb.pbuf   = fb_text;
		dpy->fb.zbuf   = NULL;
		dpy->fb.ops    = &text16_color_io;
		dpy->driver_data=text16_data_realloc(dpy->driver_data);
		if (dpy->driver_data==NULL)
			INTERNAL_ERROR;
		break;
	default:
		dpy->set_font=NULL;
		dpy->set_origin=et4000_set_origin;
		dpy->set_split=/* !!! et4000_set_split */ NULL;
		dpy->set_frame=et4000_set_frame;
		dpy->show_cursor=NULL;
		dpy->hide_cursor=NULL;
		dpy->undo_cursor=NULL;
		dpy->show_pointer=NULL;
		dpy->hide_pointer=NULL;
		dpy->undo_pointer=NULL;
		dpy->fb.dpy = dpy;
		dpy->fb.size.x = 0;
		dpy->fb.size.y = 0;
		dpy->fb.bpp    = 0;
		dpy->fb.pbuf   = NULL;
		dpy->fb.zbuf   = NULL;
		dpy->fb.ops    = NULL;
		kgim_accel_set_mode(dpy,mode);
	}
	if(chip!=ET4000) {
		/* initialize hardware pointer */
		w32_init_hw_pointer(dpy);
		dpy->show_pointer=w32_show_pointer;
		dpy->hide_pointer=w32_hide_pointer;
		dpy->undo_pointer=NULL;
	}
	memcpy(&(dpy->mode),mode,sizeof(struct kgi_mode));
}

/*
 * Takes a mode and a bits-per-pixel value
 */
static void chipset_setup_mode(struct kgi_mode *mode)
{
	int bpp=0;
	/* int fbsize=0; */ /* Not needed anymore, see below */
	
	/* By default, setup for graphics modes...
	 */
	
	switch (mode->request.graphtype) {
	case KGIGT_TEXT16:
		mode->fb     = fb_text;
		mode->fb_size= fb_text_len;
		mode->fsize  = mode->request.virt.x * mode->request.virt.y * 2;
		mode->mem    = &text16_color_mem;
		mode->io     = &text16_color_io;
		mode->mmio   = &mm_text16;
		return; /* Done setting up... */
	case KGIGT_8BIT:
		bpp=8;
		break;
	case KGIGT_15BIT:
	case KGIGT_16BIT:
		bpp=16;
		break;
	case KGIGT_24BIT:
		bpp=24;
		break;
	case KGIGT_32BIT:
		bpp=32;
		break;
	default:
		INTERNAL_ERROR;
		return;
	}

	if(w32_linear_mem.base) {
		mode->mmio = &mm_linear_graphics;
	} else {
		mode->mmio = &mm_graphics;
	} 

	mode->mmio->size  = chipinfo.ramsize;
/* this calculates instead size the same way as libggi does - OBSOLETE ? */
/*	fbsize = (mode->request.virt.x*mode->request.virt.y
		  *(bpp/8));
	fbsize = ((fbsize >> 16) + ((fbsize & 0xffff) ? 1 : 0)) << 16;
	mode->mmio->size  = fbsize;
*/
	mode->fb         = fb_graph;
	mode->fb_size    = fb_graph_len;
	mode->fsize      = 
		mode->request.virt.x * mode->request.virt.y * bpp / 8; /* was: bpp/8 */
	mode->mem        = NULL;
	mode->io         = NULL;
}

static int chipset_get_reset_mode(struct kgi_display *dpy)
{
	kgi_graphtype gmode=KGIGT_TEXT16;
	int magx=1,magy=1;
	int DCLKdiv=8;
	int dotclock;
	struct kgi_timing tx,ty;

	save_state(&reset_state);

	if (reset_state.ATC[0x10] & AR10_GRAPHICS) {
		switch (reset_state.GRC[0x05] & GR5_WRITEMODE3) {
		case GR5_WRITEMODE0:
			gmode=KGIGT_8BIT; /* 8 bit mode */
			break;
		case GR5_WRITEMODE1:
			gmode=KGIGT_1BIT; /* 1 bit mode */
			break;
		case GR5_WRITEMODE3:
			gmode=KGIGT_4BIT; /* 4 bit mode */
			break;
		case GR5_WRITEMODE2:
			break;
		}
		magy=(reset_state.CRT[0x09] & CR09_CHARHEIGHT_MASK)+1;
	} else { /* Text mode */
		gmode=KGIGT_TEXT16;
		DCLKdiv=(reset_state.SEQ[0x01] & SR01_8DOT_CHCLK) ? 8 : 9;
	}
	
	/* Get Horizontal timing */
	tx.width = (reset_state.CRT[0x01]+1) * DCLKdiv;
	tx.blankstart = reset_state.CRT[0x02] * DCLKdiv;
	tx.syncstart = reset_state.CRT[0x04] * DCLKdiv;
	tx.syncend = (reset_state.CRT[0x05] & CR05_HSE_MASK) * DCLKdiv;
	while (tx.syncend < tx.syncstart)
		tx.syncend += 0x20 * DCLKdiv;
	tx.blankend = ((reset_state.CRT[0x03] & CR03_HBE_MASK) |
	       ((reset_state.CRT[0x05] & CR05_HBLANKEND_B5) ? 0x20 : 0) |
	       (reset_state.CRT[0x01] & 0xC0)) * DCLKdiv;
	tx.total = (reset_state.CRT[0x00]+5) * DCLKdiv;
	tx.polarity = (reset_state.MISC & MISC_NEG_HSYNC)!=0;

	/* Get Vertical timing */
	ty.width = 1 + (reset_state.CRT[0x12] | 
		((reset_state.CRT[0x07] & CR07_VDISPLAYEND_B8) ? 0x100 : 0) |
		((reset_state.CRT[0x07] & CR07_VDISPLAYEND_B9) ? 0x200 : 0));
	ty.blankstart = 1 + (reset_state.CRT[0x15] |
		((reset_state.CRT[0x07] & CR07_VBLANKSTART_B8) ? 0x100 : 0) |
		((reset_state.CRT[0x09] & CR09_VBLANKSTART_B9) ? 0x200 : 0));
	ty.syncstart = (reset_state.CRT[0x10] |                
		((reset_state.CRT[0x07] & CR07_VSYNCSTART_B8) ? 0x100 : 0) |
		((reset_state.CRT[0x07] & CR07_VSYNCSTART_B9) ? 0x200 : 0));
	ty.syncend = (reset_state.CRT[0x11] & CR11_VSYNCEND_MASK) |
                (reset_state.CRT[0x12] & ~CR11_VSYNCEND_MASK);
	while (ty.syncend < ty.syncstart)
		ty.syncend += 0x10;
	ty.blankend = (reset_state.CRT[0x16] | ((ty.width-1) & ~0xFF))+1;
	ty.total = 2+(reset_state.CRT[0x06] |
		((reset_state.CRT[0x07] & CR07_VTOTAL_B8) ? 0x100 : 0) |
		((reset_state.CRT[0x07] & CR07_VTOTAL_B9) ? 0x200 : 0));
	ty.polarity = (reset_state.MISC & MISC_NEG_VSYNC)!=0;

	dotclock=(getCLK() == MISC_25MHZ_CLK) ? 25175000 : 28322000;
   
	/* Set all the mode info that we've justr gathered....
	 */
	default_kgi_modereq.frames		=	1;
	default_kgi_modereq.dpp.x		=	DCLKdiv;
	default_kgi_modereq.dpp.y=(reset_state.CRT[0x09]&CR09_CHARHEIGHT_MASK)+1;
	default_kgi_modereq.visible.x	=	tx.width / default_kgi_modereq.dpp.x;
	default_kgi_modereq.visible.y	=	ty.width / default_kgi_modereq.dpp.y;
	default_kgi_modereq.virt.x		=	tx.width / default_kgi_modereq.dpp.x;
	default_kgi_modereq.virt.y		=	ty.width / default_kgi_modereq.dpp.y;
	default_kgi_modereq.size.x		=	0;
	default_kgi_modereq.size.y		=	0;
	default_kgi_modereq.graphtype	=	gmode;
	if (dpy->defmode==NULL)
		dpy->defmode=&default_kgi_modereq;
	else
		(*dpy->defmode)=default_kgi_modereq;

	dpy->mode.request	=	default_kgi_modereq;
	dpy->mode.x			=	tx;
	dpy->mode.y			=	ty;
	dpy->mode.magnify.x	=	magx;
	dpy->mode.magnify.y	=	magy;
	dpy->mode.dclk			=	dotclock;
	dpy->mode.clk.mul		=	1;
	dpy->mode.clk.div		=	1;
	dpy->mode.lclk.mul	=	1;
	dpy->mode.lclk.div	=	1;
	dpy->mode.chipset		=	0;
	default_kgi_mode		=	dpy->mode;

	chipset_setup_mode(&default_kgi_mode);
	chipset_setup_dpy(dpy,&(default_kgi_mode));

	return EOK;
}

int kgim_chipset_init(struct kgi_card *card,struct kgi_display *dpy)   
{
	int err;

	NOTICE(DRIVER_NAME DRIVER_REV "\n");
     
	fb_text = (mem_vaddr)0x000A0000;
	fb_text_len = 0x10000;	/* 64k framebuffer */
	fb_graph_len = 0x10000;	/* We use 64K window to paged framebuffer */
	fb_graph = fb_text;
	if ((err=io_check_region(&et4000_io)))
		return err;
	if ((err=mem_check_region(&et4000_mem_graph)))
		return err;
	if ((err=io_check_region(&w32_io)))
		return err;
    
	io_claim_region(&et4000_io);
	io_claim_region(&w32_io);
	mem_claim_region(&et4000_mem_graph);
    
	outbMISC(inb(MISCr));	/* adjust IO base       */
	io_out8(0x03, 0x3bf);	/* enable et4000 extension */
	io_out8(0xa0, 0x3d8);	/* registers */
    
	if(et4000_examine_chipset()!=EOK) {
		ERROR("Detection of ET4000 failed");
		et4000_release_regions();
		return -E(CHIPSET,UNKNOWN);
	}
	
	if(chip==ET4000_W32p) {
		/* special treatment for PCI chipset */
		if(w32p_examine_chipset()!=EOK) {
			ERROR("Detection of W32p failed");
			et4000_release_regions();
			return -E(CHIPSET,UNKNOWN);
		}
	}

	/* set some default state registers */
	default_state.CRT[0x32]=inbCRT(0x32);
	default_state.CRT[0x34]=inbCRT(0x34);
	default_state.CRT[0x37]=inbCRT(0x37);

	text16_data_init();

	if(chipset_get_reset_mode(dpy)!=EOK) {
		et4000_release_regions();
		return -E(CHIPSET, UNKNOWN);
	}

	strcpy(card->manufact,chipinfo.manufact);
	strcpy(card->model,chipinfo.model);
	card->modes=chipinfo.modes;
	card->flags=chipinfo.flags;
	card->maxdots=chipinfo.maxdots;    
	card->chipset=&chipinfo;
	card->preset=default_kgi_mode;
   
	return EOK;
} 

void kgim_chipset_done(struct kgi_display *dpy)
{
	et4000_release_regions();
}

static void adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == KGIGT_TEXT16)
		? mode->request.dpp.x : 8;
    
#define ADJUST(X) mode->x.X -= (mode->x.X % foo)
    
	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);
	if (mode->x.total % foo)
		mode->x.total += foo - (mode->x.total % foo);
    
	if (((mode->x.syncend / foo) & 0x3F) == ((mode->x.syncstart / foo) & 0x3F))
		mode->x.syncend -= foo;
	if (((mode->x.blankend / foo) & 0x7F) == ((mode->x.blankstart / foo) & 0x7F))
		mode->x.blankend -= foo;
    
	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F))
		mode->y.syncend--;
	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF))
		mode->y.blankend--;
    
#undef ADJUST
}


kgi_sint	kgim_chipset_check_mode
(struct kgi_display *dpy,
 struct kgi_mode *mode,enum kgi_tm_cmd *cmd)
{
	int bpp = 0;
	int textx,texty;
	textx = mode->request.dpp.x;
	texty = mode->request.dpp.y;

	switch (mode->request.graphtype) {
		
	case KGIGT_32BIT:
		bpp=32;
		break;

#ifdef USE_UNTESTED_CODE
	case KGIGT_24BIT:
		bpp=24;
		break;
#endif

	case KGIGT_16BIT:
	case KGIGT_15BIT:
		bpp=16;
		break;
		
	case KGIGT_8BIT:
		bpp = 8; /* 1 Byte per pixel */
		break; 	       

	case KGIGT_TEXT16:
		break;	/* Not used in textmode */

	default:
		DEBUG0("graphtype %i not supported by chipset.", \
		       mode->request.graphtype);
		return -E(CHIPSET,MODE_CHECK);
	}
	
#define XVIRTUAL (mode->request.virt.x)
#define YVIRTUAL (mode->request.virt.y)
#define XVISIBLE (mode->request.visible.x)
#define YVISIBLE (mode->request.visible.y)

	switch (*cmd) {
	case CMD_PROPOSE:
		if (mode->request.graphtype == KGIGT_TEXT16) {
			/* mode->request.graphtype != KGIGT_TEXT16 */
						
			mode->magnify.x = mode->magnify.y = 1;
			if ((textx < 8)
			    || (9 < textx)) {
				DEBUG0("invalid char width (%i).",
				       textx);
				return -EINVAL;
			}
			
			if (XVIRTUAL==0 && YVIRTUAL==0)
				return -EINVAL;
			
			if (XVIRTUAL==0) 
				XVIRTUAL = fb_text_len / (YVIRTUAL * 2);
			
			if (YVIRTUAL==0) 
				YVIRTUAL = fb_text_len / (XVIRTUAL * 2);
			
			if ((2 * XVIRTUAL * YVIRTUAL) > fb_text_len) {
				DEBUG0("Not enough text memory.");
				return -ENOMEM;
			}
			mode->x.width = mode->request.visible.x * textx;
			mode->y.width = mode->request.visible.y * texty;
			
		} else {	

			if (mode->request.dpp.x == KGI_AUTO)
				mode->request.dpp.x = ggi_auto_textmode.dpp.x;
			if (mode->request.dpp.y == KGI_AUTO)
				mode->request.dpp.y = ggi_auto_textmode.dpp.y;
			
			if (XVIRTUAL == KGI_AUTO && YVIRTUAL == KGI_AUTO) {
				XVIRTUAL = (XVISIBLE == KGI_AUTO) ? 
					ggi_auto_textmode.virt.x : XVISIBLE;
				YVIRTUAL = (YVISIBLE == KGI_AUTO) ? 
					ggi_auto_textmode.virt.y : YVISIBLE;
			}
			
			if (XVISIBLE == KGI_AUTO && YVISIBLE == KGI_AUTO) {
				XVISIBLE = ggi_auto_textmode.visible.x;
				YVISIBLE = ggi_auto_textmode.visible.y;
			}
			
			if (XVIRTUAL==0 && YVIRTUAL==0)
				return -EINVAL;
			
			if (XVIRTUAL==0) 
				XVIRTUAL = (chipinfo.ramsize / bpp * 8)
					/ YVIRTUAL;
			
			if (YVIRTUAL==0) 
				YVIRTUAL = (chipinfo.ramsize / bpp * 8)
					/ XVIRTUAL;
			
			mode->magnify.x = 1;
			mode->magnify.y = 1;

			while ((XVISIBLE * mode->magnify.x) < 640) 
				mode->magnify.x++;
			
			while ((YVISIBLE * mode->magnify.y) < 350) 
				mode->magnify.y++;
			
			if (mode->magnify.y > 2)
				/* Can't do more zoom than this.. */
				return -EINVAL;
			
			if (mode->magnify.x > 2)
				/* Can't do more than doublescan... */
				return -EINVAL;

			if(XVIRTUAL <= 320) XVIRTUAL = 320; else
			if(XVIRTUAL <= 640) XVIRTUAL = 640; else
			if(XVIRTUAL <= 720) XVIRTUAL = 720; else
			if(XVIRTUAL <= 800) XVIRTUAL = 800; else
			if(XVIRTUAL <=1024) XVIRTUAL =1024; else
			if(XVIRTUAL <=1152) XVIRTUAL =1152; else
			if(XVIRTUAL <=1280) XVIRTUAL =1280; else
				{
					DEBUG0("too many horiz. pixels (%i)", XVIRTUAL);
					return -E(CHIPSET,RESOLUTION_HIGH);
				}

			if ((XVIRTUAL * YVIRTUAL * bpp / 8) > chipinfo.ramsize)	{
					if ((XVISIBLE * YVISIBLE * bpp / 8) > chipinfo.ramsize) {
						DEBUG0("not enough memory to" \
						       " do virtual %ix%i at %i bpp", \
						       XVIRTUAL, YVIRTUAL, bpp);
						return -ENOMEM;
					} 
					XVIRTUAL=XVISIBLE;
					YVIRTUAL=YVISIBLE;
			}
			mode->request.visible.x &= 0x7FFFFFF8;
			mode->request.visible.y &= 0x7FFFFFF8;
  			mode->x.width = mode->request.visible.x * mode->magnify.x;
			mode->y.width = mode->request.visible.y * mode->magnify.y;
		}
		
		mode->clk.div=mode->lclk.div=1;
		return EOK;
		
	case CMD_LOWER:
		
		if (mode->dclk*mode->clk.mul/mode->clk.div < chipinfo.dclk.min) {
			DEBUG0("DCLK too low (%i).", mode->dclk);
			return -E(CHIPSET,CLK_LOW_LIMIT);
		}
		mode->dclk=(mode->dclk*mode->clk.mul/mode->clk.div > chipinfo.dclk.max)
			? chipinfo.dclk.max / mode->clk.mul * mode->clk.div : mode->dclk;
		
		adjust_timing(mode);
		return EOK;
		
	case CMD_RAISE:
		
		if (mode->dclk*mode->clk.mul/mode->clk.div > chipinfo.dclk.max) {
			DEBUG0("DCLK too high (%i).", \
			       mode->dclk);
			return -E(CHIPSET,CLK_HIGH_LIMIT);
		}
		mode->dclk=(mode->dclk*mode->clk.mul/mode->clk.div < chipinfo.dclk.min)
			? chipinfo.dclk.min / mode->clk.mul * mode->clk.div : mode->dclk;
		
		adjust_timing(mode);
		return EOK;
		
	case CMD_CHECK:
		switch (mode->lclk.div) {
		case 1:
			break;
		default:
			DEBUG0("invalid CLK to LCLK divider %d.",
			       mode->lclk.div);
			return -E(CHIPSET,MODE_CHECK);
		} 
		if (mode->request.graphtype == KGIGT_TEXT16) {
			mode->flags |= FL_CHIP_COLORTEXT;
			if ((mode->magnify.x != 1) || (mode->magnify.y != 1)	/*driver */
			    ||(textx < 8)	/* hardware */
			    ||(9 < textx)	/* hardware */
			    ||(texty < 1)	/* hardware */
			    ||(texty > 32)	/* hardware */
			    ||((2 * XVIRTUAL * YVIRTUAL) > fb_text_len)
			    || (mode->dclk > chipinfo.dclk.max)		/* hardware */
			    ||(mode->dclk < chipinfo.dclk.min)) {	/* hardware */
				ERROR("hw check failed");
				return -E(CHIPSET,MODE_CHECK);
			}
		} else {
			switch(XVIRTUAL) {
			case 320:
			case 640:
			case 720:
			case 800:
			case 1024:
			case 1152:
			case 1280:
				break;
			default:
				return -E(CHIPSET,MODE_CHECK);
			}

			if (((XVIRTUAL * YVIRTUAL * bpp / 8) > chipinfo.ramsize)
			    || (mode->magnify.x > 2) || (mode->magnify.y > 2)
			    || (mode->dclk*mode->clk.mul/mode->clk.div < chipinfo.dclk.min)
			    || (chipinfo.dclk.max < mode->dclk*mode->clk.mul/mode->clk.div)
			    || (mode->magnify.x==2 && mode->request.graphtype != KGIGT_8BIT)
			    || (mode->request.virt.x >= 1024 &&
				mode->request.graphtype == KGIGT_32BIT)) {
				ERROR("hw check (graphics) failed");
				return -E(CHIPSET,MODE_CHECK);
			}
		}
		
#define TM(X) (((unsigned) chipinfo.maxdots.x < mode->x.X)\
		|| ((unsigned) chipinfo.maxdots.y < mode->y.X))
		if (TM(width) || TM(blankstart) || TM(syncstart)
		|| TM(syncend) || TM(blankend) || TM(total)) {
		TRACE(printk(KERN_ERR "%s: timing check failed\n", __FILE__));
		TRACE(printk(KERN_ERR "%s: Maxdots are (%d,%d)\n", __FILE__,(unsigned)chipinfo.maxdots.x,(unsigned) chipinfo.maxdots.y));
		return -E(CHIPSET,MODE_CHECK);
	}
#undef TM
		
		chipset_setup_mode(mode);
		return EOK;
	
	default:
		INTERNAL_ERROR;
		return -E(CHIPSET,FAILED);
	}
    
#undef XVIRTUAL
#undef YVIRTUAL
#undef XVISIBLE
#undef YVISIBLE
}

void kgim_chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	struct statetable setup = default_state;
	register int foo;
	int DCLKdiv, mul;
	int textx,texty;
	textx=mode->request.dpp.x;
	texty=mode->request.dpp.y;


	/*
	        Setup bits per pixel and logical screen width
	*/

	switch (mode->request.graphtype) {
	case KGIGT_TEXT16:
		foo = 2 * mode->request.virt.x;
		foo /= 4; /* due to the addressing mode used. */
		break;

	case KGIGT_8BIT:
		foo = mode->request.virt.x >> 3;
		break;

	case KGIGT_15BIT:
	case KGIGT_16BIT:
		foo = (mode->request.virt.x >> 3) * 2;
		break;

	case KGIGT_24BIT:
		foo = (mode->request.virt.x >> 3) * 3;
		break;

	case KGIGT_32BIT:
		foo = (mode->request.virt.x >> 3) * 4;
		break;

	default:
		INTERNAL_ERROR;
		return;
	}
	
	setup.CRT[CRT_LOGICALWIDTH] |= foo & 0xFF;
	setup.CRT[CRT_OVRFLW_H] |= (foo & 0x100) ? CR3F_LOGICALWIDTH_B8 : 0;

	if (mode->request.graphtype == KGIGT_TEXT16) {
		setup.CRT[0x17] = CR17_CGA_ADDR_WRAP
			| CR17_ENABLE_SYNC | CR17_CGA_BANKING
			| CR17_HGC_BANKING;
		
		setup.CRT[0x14] |= CR14_UNDERLINE_MASK;
		
		setup.SEQ[0x02] = SR02_PLANE0 | SR02_PLANE1;
		setup.SEQ[0x07] |= SR07_DIV2;

		setup.GRC[0x05] |= GR5_ODD_EVEN;
		setup.GRC[0x06] |= GR6_CHAIN_OE;
		setup.GRC[0x06] |= GR6_MAP_A0_64;
		
		switch ((DCLKdiv = textx)) {
		case 8:
			setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
			break;
		case 9:
			setup.ATC[0x10] |= AR10_9DOT_CHARS;
			break;
		default:
			INTERNAL_ERROR;
			return;
		}
		setup.CRT[0x09] |= texty - 1;
		setup.ATC[0x13] = (textx == 9) ? 0x08 : 0x00;
	} else { /* mode->request.graphtype == KGIGT_TEXT16 */
		DCLKdiv = 8;
		
		for (foo = 0; foo < 16; foo++)
			setup.ATC[foo] = foo;
		
		setup.SEQ[0x02] |= SR02_PLANE_MASK;
		setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
		setup.SEQ[0x02] |= 0x0f;
		setup.SEQ[0x04] |= SR04_CHAINED | SR04_NO_ODDEVEN 
			| SR04_256K_ACCESS; 		

		setup.ATC[0x10] |= AR10_GRAPHICS;
		
		setup.GRC[0x06] |= GR6_GRAPHMODE | GR6_MAP_A0_64;
		setup.GRC[0x05] |= GR5_8BIT_COLOR;
		setup.GRC[0x08] |= 0xff; /* Enable all bits */
		
		setup.CRT[0x09] |= (mode->magnify.y-1) & CR09_CHARHEIGHT_MASK;
		setup.CRT[0x14] |= CR14_DOUBLEWORDMODE; 

		if(chip==ET4000_W32p) {
			/* switch on linear addressing */
			setup.CRT[0x36] |= 0x10;
			setup.CRT[0x30] = (law_base >> 22) & 0xFF;
		}

		switch (mode->request.graphtype) {
		case KGIGT_8BIT:
			if(mode->request.virt.x > 320)
  			{ if (chip==ET4000) 
			  { setup.ATC[0x16] |= AR16_HIRES_8BIT; /* special for "old" ET4000s */ 
			    setup.ATC[0x10] |= AR10_8BIT_MODE;
			    setup.SEQ[0x07] |= SR07_DIV2;
			  }
			  else /* W32 cards don't need above (also ET6000.c) */
			  {  setup.CRT[0x17] |= CR17_WORDMODE;
			     setup.CRT[0x14] |= CR14_MEMCOUNT_BY_4;
			  }
			}
			else /* just "normal" VGA-modes */
			{  setup.SEQ[0x07] |= SR07_DIV2;
			   setup.ATC[0x10] |= AR10_8BIT_MODE;
			}
			break;
/*
#if USE_UNTESTED_CODE
		case KGIGT_24BIT:
			setup.ATC[0x10] |= AR10_8BIT_MODE;
			if(mode->request.virt.x > 320)
				setup.ATC[0x16] |= AR16_HIRES_8BIT;
			setup.SEQ[0x07] |= SR07_DIV2;
			break;
#endif
*/
		case KGIGT_15BIT:
		case KGIGT_16BIT:
		case KGIGT_24BIT:
		case KGIGT_32BIT:
			if(chip==ET4000) {
				setup.ATC[0x16] |= AR16_HIRES_16BIT;
			} else {
				setup.ATC[0x16] |= AR16_W32_HICOLOR;
			}
			setup.CRT[0x14] |= CR14_MEMCOUNT_BY_4; 
			break;

		default:
			INTERNAL_ERROR;
			return;
		}
	}
	
	/*
	        setup horizontal timing
	*/

	mul = mode->lclk.mul;

	foo = (mode->x.width / DCLKdiv * mul) - 1; /* horizontal display end */
	setup.CRT[0x01] |= foo & 0xFF;

	foo = (mode->x.blankstart / DCLKdiv * mul); /* horizontal blank start */
	setup.CRT[0x02] |= foo & 0xFF;
	setup.CRT[0x3F] |= (foo & 0x100) ? CR3F_HBLANKSTART_B8 : 0;

	foo = (mode->x.syncstart / DCLKdiv * mul); /* horizontal sync start  */
	setup.CRT[0x04] |= foo & 0xFF;
	setup.CRT[0x3F] |= (foo & 0x100) ? CR3F_HSYNCSTART_B8 : 0;

	foo = (mode->x.syncend / DCLKdiv * mul); /* horizontal sync end    */
	setup.CRT[0x05] |= foo & CR05_HSE_MASK;

	foo = (mode->x.blankend / DCLKdiv * mul); /* horizontal blank end   */
	setup.CRT[0x03] |= foo & CR03_HBE_MASK;
	setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;

	foo = (mode->x.total / DCLKdiv * mul) - 5; /* horizontal total       */
	setup.CRT[0x00] |= foo & 0xFF;
	setup.CRT[0x3F] |= (foo & 0x100) ? CR3F_HTOTAL_B8 : 0;

	/*
	        setup vertical timing
	*/

	foo = mode->y.width - 1; /* vertical display end   */
	setup.CRT[0x12] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x400) ? CR35_VDISPLAYEND_B10 : 0;

	foo = mode->y.blankstart - 1; /* vertical blank start   */
	setup.CRT[0x15] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8 : 0;
	setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x400) ? CR35_VBLANKSTART_B10 : 0;

	foo = mode->y.syncstart; /* vertical sync start    */
	setup.CRT[0x10] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x400) ? CR35_VSYNCSTART_B10 : 0;

	foo = mode->y.syncend;	/* vertical sync end      */
	setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;

	foo = mode->y.blankend + 1; /* end vertical blank     */
	setup.CRT[0x16] |= foo & 0xFF;

	foo = mode->y.total - 2; /* vertical total         */
	setup.CRT[0x06] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9 : 0;
	setup.CRT[0x35] |= (foo & 0x400) ? CR35_VTOTAL_B10 : 0;

	setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
	setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;

	/* set splitline to 0 */
	setup.CRT[0x18] = 0;
	setup.CRT[0x07] &= ~CR07_LINECOMPARE_B8;
	setup.CRT[0x09] &= ~CR09_LINECOMPARE_B9;
	setup.CRT[0x35] &= ~CR35_LINECOMPARE_B10;

	/*
	        Ok. We'll do the programming now.
	*/

#ifdef DEBUG
	printk_mode(mode,"Set mode:");
#endif /* DEBUG */

	kgim_monitor_set_mode(dpy,mode);
	kgim_ramdac_set_mode(dpy,mode);
	kgim_clock_set_mode(dpy,mode);
	restore_state(&setup);

	chipset_setup_dpy(dpy,mode);

	select_segment(0);	/* reset to segment 0 */

	DEBUG1("Mode set complete");
}

int kgim_chipset_command(struct kgi_graphic *state,kgi_uint cmd,kgi_uint arg)
{
	kgi_coord  *coord;
	struct kgi_display *dpy;
	
	dpy = state->dev.dpy;

	switch (cmd) {
	case CHIP_SETSPLITLINE:
		if (arg > state->dev.mode.virt.y)
			return -E(CHIPSET,BOGUS);
		
		kgi_set_split(&(state->dev),arg);
		return 0;
	case CHIP_SETVISFRAME:
		coord = (kgi_coord *)(state->io_buf);
		
		if (coord->x < 0 || 
		    coord->x > state->dev.mode.virt.x - state->dev.mode.visible.x)
			return -E(CHIPSET,BOGUS);
		
		if (coord->y < 0 || 
		    coord->y > state->dev.mode.virt.y - state->dev.split/dpy->mode.magnify.y)
			return -E(CHIPSET,BOGUS);
		
		kgi_set_origin(&(state->dev),0,coord->x,coord->y);
		return 0;
	case CHIP_SETPAGE:
		if(chip>=ET4000_W32) {
			if(arg & ~0x3F)	
				return -E(CHIPSET, BOGUS);
		} else {
			if(arg & ~0x0F)
				return -E(CHIPSET, BOGUS);
		}
		select_segment(arg);
		return 0;
	default:
		break;
	}
	
	return -E(CHIPSET,UNKNOWN);
}
